from dependencies.core import torchvision
from libs import ImageToMatPlotLib

def flowers_unshaped():
  return {
    "valid": torchvision.transforms.Compose([
      torchvision.transforms.RandomRotation(21),
      torchvision.transforms.RandomPerspective(),
      torchvision.transforms.Resize(228),
      torchvision.transforms.CenterCrop(228),

      torchvision.transforms.ToTensor(),
      torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ]),
    "display": torchvision.transforms.Compose([
      torchvision.transforms.ToTensor(),
      ImageToMatPlotLib((-1, )),
    ]),
    "test": torchvision.transforms.Compose([
      torchvision.transforms.Resize(228),
      torchvision.transforms.CenterCrop(228),

      torchvision.transforms.ToTensor(),
      torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ]),
    "train": torchvision.transforms.Compose([
      torchvision.transforms.ColorJitter(),
      torchvision.transforms.RandomAffine(42),
      torchvision.transforms.RandomHorizontalFlip(),
      torchvision.transforms.RandomRotation(32),
      torchvision.transforms.RandomVerticalFlip(),
      torchvision.transforms.RandomRotation(21),
      torchvision.transforms.RandomPerspective(),
      torchvision.transforms.Resize(228),
      torchvision.transforms.CenterCrop(228),

      torchvision.transforms.ToTensor(),
      torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])
  }



from dependencies.core import torch
from dependencies.datatypes import Parameter, Iterator
from modules.models.abs_module import AbsModule

class Elemental(AbsModule):
  """
    A elemental neural network module, only for testing purposes.

    Attributes:
    -----------
      labels: list
        A list of labels.
      features: list
        A list of features.
      input_size: int
        The input size.
      num_classes: int
        The number of classes.

    Methods:
    --------
      forward(self, x: torch.Tensor) -> torch.Tensor
        A forward pass of the neural network.
      load(self, path: str) -> None
        A load of the neural network.
      save(self, path: str) -> str
        A save of the neural network.
      create(self, input_size: int, num_classes: int) -> None
        Re creates the neural network.
  """

  def __repr__(self):
    """
      A string representation of the neural network.

      Returns:
      --------
        : str
          A string representation 'Basic()'.
    """
    return 'Elemental()'

  def __init__(self, *, data: dict|None = None, labels:list = [], input_size: int = 28, dropout: float = 0.0,
               num_classes: int|None = None) -> None:
    """
      Constructor of the neural network.

      Parameters:
      -----------
        data: dict
          A dictionary containing the data, to load the network though the pth file.
        labels: list
          A list of labels.
        input_size: int
          The input size.
        dropout: float
          The dropout probability.
    """
    if data is None:
      AbsModule.__init__(self,
        labels = labels, input_size = input_size, dropout = dropout, num_classes = num_classes
      )
      self.features = torch.nn.Sequential(
        torch.nn.Linear(self.input_size, self.num_classes),
        torch.nn.LogSoftmax(dim=1)
      )
    else:
      AbsModule.__init__(self, data = data)

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor

      Help:
      -----
        model.forward = lambda x: model.classifier(model.features(x)).view(x.size(0), class_count)
    """
    x = self.features(x)
    return torch.flatten(x, 1)

  def parameters(self) -> Iterator[Parameter]:
    """
      Get the parameters of the neural network.

      Returns:
      --------
        Iterator[Parameter]
    """
    return self.features.parameters()

  def create(self, *, input_size: int = 28, dropout: float|None = None) -> None: # pyright: ignore
    """
      Re creates the neural network.

      Parameters:
      -----------
        input_size: int
          The input size.
    """



from dependencies.core import click, inspect, Image
from .shared import labels_reader
from modules import models
MODELS_NAMES = [ name for name, _ in inspect.getmembers(models, predicate=inspect.isclass) ]

@click.command()
@click.pass_context
@click.argument("architecture", default=MODELS_NAMES[0], type=click.Choice(MODELS_NAMES), required=1)
@click.argument("categories-file", type=click.Path(file_okay=True, writable=True), required=1)
@click.argument("save-file", type=click.Path(file_okay=True, writable=True), required=1)
@click.option('-d', '--dropout', type=click.FLOAT, default=0.0)
@click.option('-a', '--from-image', type=click.Path(file_okay=True, readable=True),
  help="set the input-size from an image, resulting in {width X height}"
)
def create(_, architecture: str, categories_file: str, save_file: str,
           dropout: float, input_size: int, from_image: str) -> None:
  """
    Creates a new model for a given architecture.
    This will generate a .pth file to use later to train, test, and evaluate the model.
  """
  if from_image:
    img = Image.open(from_image)
    input_size = img.size[0] * img.size[1] * len(img.split()) # Check channels, usually L|RGB|RGBA
  sorted_labels = labels_reader(categories_file)
  kwargs = { "labels": sorted_labels, "input_size": input_size }

  if dropout > 0.0:
    kwargs["dropout"] = dropout

  model = getattr(models, architecture)(**kwargs)
  model.save(save_file)
  print('💚')



from dependencies.core import torch, abstract_attribute, ABCMeta, textwrap, sys, time
from dependencies.datatypes import Parameter, Iterator

# TODO: Move this as a library
def sizeof_formated(num, suffix='B'):
    """
      Converts a size in bytes to a human-readable format.

      Parameters:
      -----------
        num: int
          Size in bytes.
        suffix: str
          Suffix to append to the unit.

      Returns:
      --------
        : str
          Human-readable size.
    """
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Yi{suffix}"

class AbsModule(torch.nn.Module, metaclass=ABCMeta):
  @abstract_attribute
  def labels(self) -> list|None:
    pass

  def __call__(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor
    """
    return self.forward(x)

  def __str__(self):
    """
      A verbose string representation of the neural network.

      Returns:
      --------
        : str
          labels, features, classifier
    """
    features_str = "\n  ".join(str(self.features).split("\n"))
    labels_str = textwrap.fill(str(self.labels)[1:-1], width=80)
    labels_str = labels_str.replace('\n', '\n' + ' '*4)
    formated_size = sizeof_formated(sys.getsizeof(self))

    if hasattr(self, "classifier"):
      classifier_str = "\n  ".join(str(self.classifier).split("\n"))

    if hasattr(self, 'dropout'):
      meta_str = f"size: {formated_size},\tdropout: {self.dropout},\tinput size:{self.input_size}" + \
        f"\t label count (output): {self.num_classes}"
    else:
      meta_str = f"size: {formated_size},\tinput size:{self.input_size}," + \
        f"\t label count (output): {self.num_classes}"

    training_history_str = ""*4
    for line in self.training_history: # pyright: reportGeneralTypeIssues=false
      training_history_str += f"Epochs: {line[0][1]}, "
      training_history_str += f"Batch: {line[1][1]}, "
      training_history_str += f"Items: ({line[2][1]}, {line[2][2]})\n"

      time_f = time.strftime("%H:%M:%S", time.gmtime(line[3][1]))
      training_history_str += ' '*6 + f"Accuracy: {line[4][1]:.2f}%, "
      training_history_str += f"Time: {time_f}\n" + " "*4
    training_history_str = training_history_str[:-(4-1)] # 4 will be a constant

    return f"{self.__repr__()} (\n" + \
      f"  (meta): \n    {meta_str}\n" + \
      f"  (labels): \n    {labels_str}\n" + \
      f"  (features): {features_str}\n" + \
      (f"  (classifier): {classifier_str}\n" if hasattr(self, 'classifier') else '') + \
      (f"  (training history): \n    {training_history_str[:-1]}" if len(self.training_history) > 0 else '') + \
    f")"

  def parameters(self) -> Iterator[Parameter]:
    """
      Get the parameters of the neural network.

      Returns:
      --------
        Iterator[Parameter]
    """
    return self.features.parameters()

  def save(self, path: str) -> None:
    """
      Save the neural network.

      Parameters:
      -----------
        path: str
          path to save model

      Returns:
      --------
        None
    """
    obj = {
      'name': self.__repr__()[:-2], # Trim the () ex. Basic()/Basic
      'input_size': self.input_size,
      'labels': self.labels,
      'features': self.features,
      'state_dict': self.state_dict(),
    }

    if hasattr(self, 'training_history'):
      obj['training_history'] = self.training_history
    if hasattr(self, 'dropout') and self.dropout > 0.0:
      obj['dropout'] = self.dropout
    if hasattr(self, 'classifier'):
      obj['classifier'] = self.classifier
    if hasattr(self, 'avgpool'): # Supporting AlexNet
      obj['avgpool'] = self.avgpool

    torch.save(obj, path)

  def modify(self, *, labels:list|None = None, input_size: int|None = None, num_classes: int|None = None,
             dropout: float|None = None, state_dict: dict|None = None) -> None:
    """
      Parameters:
      -----------
        url: str|None
          The url to load the model from.
        state_dict: dict
          The state dict to load the model from.

      Returns:
      --------
        None
    """
    if labels is not None:
      self.labels = labels
    if num_classes is not None:
      self.num_classes = num_classes
    if input_size is not None:
      self.input_size = input_size
    if dropout is not None:
      self.dropout = dropout
    if state_dict is not None:
      self.load_state_dict(state_dict)

  def __init__(self, *, data: dict|None = None, labels:list = [], input_size: int = 28, dropout: float = 0.0,
               num_classes: int|None = None) -> None:
    """
      Constructor of the neural network.

      Parameters:
      -----------
        data: dict
          A dictionary containing the data, to load the network though the pth file.
        labels: list
          A list of labels.
        input_size: int
          The input size.
        dropout: float
          The dropout probability.
    """
    super().__init__()
    if data is None:
      self.labels = labels
      self.num_classes = len(labels) if num_classes is None else num_classes
      self.training_history = []
      self.input_size = input_size
      self.dropout = dropout
    else:
      self.labels = data['labels']
      self.num_classes = len(self.labels)
      self.input_size = data['input_size']
      self.features = data['features']

      if 'dropout' in data:
        self.dropout = data['dropout']
      if 'training_history' in data:
        self.training_history = data['training_history']
      if 'dropout' in data:
        self.dropout = data['dropout']
      if 'classifier' in data:
        self.classifier = data['classifier']
      if 'avgpool' in data:
        self.avgpool = data['avgpool']



from dependencies.core import torch
from .abs_module import AbsModule
from torchvision.models import SqueezeNet # Alias

class Squeezenet(SqueezeNet, AbsModule):
  """
    A basic neural network module. With randomly selected features.

    Attributes:
    -----------
      labels: list
        A list of labels.
      features: list
        A list of features.
      classifier: torch.nn.Module
        A classifier.
      input_size: int
        The input size.
      num_classes: int
        The number of classes.
      dropout: float
        The dropout probability.

    Methods:
    --------
      forward(self, x: torch.Tensor) -> torch.Tensor
        A forward pass of the neural network.
      load(self, path: str) -> None
        A load of the neural network.
      save(self, path: str) -> str
        A save of the neural network.
      create(self, input_size: int, num_classes: int, dropout: float) -> None
        Re creates the neural network.
  """

  def __repr__(self):
    """
      A string representation of the neural network.

      Returns:
      --------
        : str
          A string representation 'Basic()'.
    """
    return 'Squeezenet()'

  def __init__(self, *, data: dict|None = None, labels:list = [], input_size: int = 28, dropout: float = 0.0,
               num_classes: int|None = None) -> None:
    """
      Constructor of the neural network.

      Parameters:
      -----------
        data: dict
          A dictionary containing the data, to load the network though the pth file.
        labels: list
          A list of labels.
        input_size: int
          The input size.
        dropout: float
          The dropout probability.
    """
    if num_classes is None:
      num_classes = len(labels)

    if data is None:
      if dropout > 0.0:
        AbsModule.__init__(self,
          labels = labels, input_size = input_size, dropout = dropout, num_classes = num_classes
        )
        SqueezeNet.__init__(self, version="1_1", num_classes = num_classes, dropout=dropout)
      else:
        AbsModule.__init__(self, labels = labels, input_size = input_size, num_classes = num_classes)
        SqueezeNet.__init__(self, version="1_1", num_classes = num_classes)
      # SqueezeNet destroys the labels attribute.
      self.labels = labels  # pyright: reportGeneralTypeIssues=false
    else:
      AbsModule.__init__(self, data = data)



from dependencies.core import torchvision
from libs import ImageToMatPlotLib, UnShapetransform, Reshapetransform

def mnist():
  return {
    "valid": torchvision.transforms.Compose([
      torchvision.transforms.Grayscale(),
      torchvision.transforms.Resize(28),
      torchvision.transforms.CenterCrop(28),
      torchvision.transforms.ToTensor(),
      torchvision.transforms.Normalize((0.5,), (0.5,)),
      Reshapetransform((-1, )),
    ]),
    "display": torchvision.transforms.Compose([
      torchvision.transforms.Grayscale(),
      torchvision.transforms.ToTensor(),
      torchvision.transforms.Normalize((0.5,), (0.5,)),
      UnShapetransform((-1, )),
      ImageToMatPlotLib((-1, )),
    ]),
    "test": torchvision.transforms.Compose([
      torchvision.transforms.Grayscale(),
      torchvision.transforms.Resize(28),
      torchvision.transforms.CenterCrop(28),
      torchvision.transforms.ToTensor(),
      torchvision.transforms.Normalize((0.5,), (0.5,)),
      Reshapetransform((-1, )),
    ]),
    "train": torchvision.transforms.Compose([
      torchvision.transforms.Grayscale(),
      torchvision.transforms.Resize(28),
      torchvision.transforms.CenterCrop(28),
      torchvision.transforms.ToTensor(),
      torchvision.transforms.Normalize((0.5,), (0.5,)),
      Reshapetransform((-1, )),
    ])
  }



from dependencies.core import click, torch, urllib, tempfile
from .shared import labels_reader
from modules import models

@click.command()
@click.pass_context
@click.argument("model_file", type=click.Path(file_okay=True, exists=True, readable=True), required=1)
@click.option('-r', '--re-create', is_flag=True, help="Re-create the model with passed parameters")
@click.option('-c', '--categories-file', type=click.Path(file_okay=True, writable=True))
@click.option('-k', '--num-classes', type=click.INT, default=None)
@click.option('-i', '--input-size', type=click.INT, default=None)
@click.option('-s', '--state-dict-file', type=click.Path(file_okay=True, writable=True))
@click.option('-w', '--state-dict-weights-url', type=click.STRING,
              help="In example: https://download.pytorch.org/models/squeezenet1_1-f364aa15.pth")
@click.option('-d', '--dropout', type=click.FLOAT, default=0.0)
@click.option('-h', '--hidden-layers', type=click.INT)
def modify(_, model_file, re_create, categories_file, num_classes, input_size, state_dict_file,
           state_dict_weights_url, dropout, hidden_layers):
  """
    Changes the hyper parameters of a model.
    It also allows you to use pre-trained weights, by re-creating the model
  """
  kwargs = {}
  if categories_file is not None:
    kwargs["labels"] = labels_reader(categories_file)
  if num_classes is not None:
    kwargs["num_classes"] = num_classes
  if input_size is not None:
    kwargs["input_size"] = input_size
  if dropout is not None:
    kwargs["dropout"] = dropout
  if hidden_layers is not None:
    kwargs["hidden_layers"] = hidden_layers
    raise NotImplementedError("Changing hidden layers is not yet implemented.")

  if state_dict_file is not None:
    kwargs["state_dict"] = torch.load(state_dict_file)
  if state_dict_weights_url is not None:
    print('💙', end='\r')
    with tempfile.NamedTemporaryFile(mode="w+") as file:
      urllib.request.urlretrieve(state_dict_weights_url, file.name) # pyright: reportGeneralTypeIssues=false
      kwargs["state_dict"] = torch.load(file.name)

  print('💛', end='\r')
  data = torch.load(model_file)
  if re_create:
    init_kwargs = {}
    for key in ['input_size', 'labels', 'num_classes', 'dropout']:
      if not key in kwargs:
        init_kwargs[key] = data[key]
      else:
        init_kwargs[key] = kwargs[key]

    model = getattr(models, data['name'])(**init_kwargs)
  else:
    model = getattr(models, data['name'])(**{"data": data})

  model.modify(**kwargs)
  model.save(model_file)
  print('💚')



# pyright: reportUnusedImport=false
from commands.create import create
from commands.download import download
from commands.info import info
from commands.predict import predict
from commands.test import test
from commands.train import train
from commands.compare import compare
from commands.modify import modify
